{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# A from scratch example on the Sellar problem\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Introduction\nIn this example,\nwe will create an MDO scenario based on the Sellar's problem from scratch.\nContrary to the `|g| in ten minutes tutorial <gemseo_10min>`,\nall the disciplines will be implemented from scratch\nby sub-classing the :class:`.MDODiscipline` class\nfor each discipline of the Sellar problem.\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## The Sellar problem\nWe will consider in this example the Sellar problem:\n\n.. include:: /tutorials/_description/sellar_problem_definition.inc\n\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Imports\nAll the imports needed for the tutorials are performed here.\nNote that some of the imports are related to the Python 2/3 compatibility.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "from __future__ import division, unicode_literals\n\nfrom math import exp\n\nfrom matplotlib import pyplot as plt\nfrom numpy import array, ones\n\nfrom gemseo.algos.design_space import DesignSpace\nfrom gemseo.api import configure_logger, create_scenario\nfrom gemseo.core.discipline import MDODiscipline\n\nconfigure_logger()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Create the disciplinary classes\nIn this section,\nwe define the Sellar disciplines by sub-classing the :class:`.MDODiscipline` class.\nFor each class,\nthe constructor and the _run method are overriden:\n\n- In the constructor,\n  the input and output grammar are created.\n  They define which inputs and outputs variables are allowed\n  at the discipline execution.\n  The default inputs are also defined,\n  in case of the user does not provide them at the discipline execution.\n- In the _run method is implemented the concrete computation of the discipline.\n  The inputs data are fetch\n  by using the :meth:`.MDODiscipline.get_inputs_by_name` method.\n  The returned NumPy arrays can then be used to compute the output values.\n  They can then be stored in the :attr:`!MDODiscipline.local_data` dictionary.\n  If the discipline execution is successful.\n\nNote that we do not define the Jacobians in the disciplines.\nIn this example,\nwe will approximate the derivatives using the finite differences method.\n\n### Create the SellarSystem class\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class SellarSystem(MDODiscipline):\n    def __init__(self):\n        super(SellarSystem, self).__init__()\n        # Initialize the grammars to define inputs and outputs\n        self.input_grammar.initialize_from_data_names([\"x\", \"z\", \"y_1\", \"y_2\"])\n        self.output_grammar.initialize_from_data_names([\"obj\", \"c_1\", \"c_2\"])\n        # Default inputs define what data to use when the inputs are not\n        # provided to the execute method\n        self.default_inputs = {\n            \"x\": ones(1),\n            \"z\": array([4.0, 3.0]),\n            \"y_1\": ones(1),\n            \"y_2\": ones(1),\n        }\n\n    def _run(self):\n        # The run method defines what happens at execution\n        # ie how outputs are computed from inputs\n        x, z, y_1, y_2 = self.get_inputs_by_name([\"x\", \"z\", \"y_1\", \"y_2\"])\n        # The ouputs are stored here\n        self.local_data[\"obj\"] = array([x[0] ** 2 + z[1] + y_1[0] ** 2 + exp(-y_2[0])])\n        self.local_data[\"c_1\"] = array([3.16 - y_1[0] ** 2])\n        self.local_data[\"c_2\"] = array([y_2[0] - 24.0])"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Create the Sellar1 class\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class Sellar1(MDODiscipline):\n    def __init__(self):\n        super(Sellar1, self).__init__()\n        self.input_grammar.initialize_from_data_names([\"x\", \"z\", \"y_2\"])\n        self.output_grammar.initialize_from_data_names([\"y_1\"])\n        self.default_inputs = {\n            \"x\": ones(1),\n            \"z\": array([4.0, 3.0]),\n            \"y_1\": ones(1),\n            \"y_2\": ones(1),\n        }\n\n    def _run(self):\n        x, z, y_2 = self.get_inputs_by_name([\"x\", \"z\", \"y_2\"])\n        self.local_data[\"y_1\"] = array(\n            [(z[0] ** 2 + z[1] + x[0] - 0.2 * y_2[0]) ** 0.5]\n        )"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Create the Sellar2 class\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class Sellar2(MDODiscipline):\n    def __init__(self):\n        super(Sellar2, self).__init__()\n        self.input_grammar.initialize_from_data_names([\"z\", \"y_1\"])\n        self.output_grammar.initialize_from_data_names([\"y_2\"])\n        self.default_inputs = {\n            \"x\": ones(1),\n            \"z\": array([4.0, 3.0]),\n            \"y_1\": ones(1),\n            \"y_2\": ones(1),\n        }\n\n    def _run(self):\n        z, y_1 = self.get_inputs_by_name([\"z\", \"y_1\"])\n        self.local_data[\"y_2\"] = array([abs(y_1[0]) + z[0] + z[1]])"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Create and execute the scenario\n\n### Instantiate disciplines\nWe can now instantiate the disciplines\nand store the instances in a list which will be used below.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "disciplines = [Sellar1(), Sellar2(), SellarSystem()]"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Create the design space\nIn this section,\nwe define the design space which will be used for the creation of the MDOScenario.\nNote that the coupling variables are defined in the design space.\nIndeed,\nas we are going to select the IDF formulation to solve the MDO scenario,\nthe coupling variables will be unknowns of the optimization problem\nand consequently they have to be included in the design space.\nConversely,\nit would not have been necessary to include them\nif we aimed to select an MDF formulation.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "design_space = DesignSpace()\ndesign_space.add_variable(\"x\", 1, l_b=0.0, u_b=10.0, value=ones(1))\ndesign_space.add_variable(\n    \"z\", 2, l_b=(-10, 0.0), u_b=(10.0, 10.0), value=array([4.0, 3.0])\n)\ndesign_space.add_variable(\"y_1\", 1, l_b=-100.0, u_b=100.0, value=ones(1))\ndesign_space.add_variable(\"y_2\", 1, l_b=-100.0, u_b=100.0, value=ones(1))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Create the scenario\nIn this section,\nwe build the MDO scenario which links the disciplines with the formulation,\nthe design space and the objective function.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "scenario = create_scenario(\n    disciplines, formulation=\"IDF\", objective_name=\"obj\", design_space=design_space\n)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Add the constraints\nThen,\nwe have to set the design constraints\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "scenario.add_constraint(\"c_1\", \"ineq\")\nscenario.add_constraint(\"c_2\", \"ineq\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "As previously mentioned,\nwe are going to use finite differences to approximate the derivatives\nsince the disciplines do not provide them.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "scenario.set_differentiation_method(\"finite_differences\", 1e-6)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Execute the scenario\nThen,\nwe execute the MDO scenario with the inputs of the MDO scenario as a dictionary.\nIn this example,\nthe gradient-based `SLSQP` optimizer is selected, with 10 iterations at maximum:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "scenario.execute(input_data={\"max_iter\": 10, \"algo\": \"SLSQP\"})"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### Post-process the scenario\nFinally,\nwe can generate plots of the optimization history:\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "scenario.post_process(\"OptHistoryView\", save=False, show=False)\n# Workaround for HTML rendering, instead of ``show=True``\nplt.show()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.8.12"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}